////
// Arduino CNC Joystick Controller
// Written by: Robert Canthal
// Feb 2019
////
//
//// Connections
//  TFT Display - connect to the following pins.    
//  Use a level shifter as the pins are only 3.3v rated on the display.
//  - CLK   - 13 - must use this pin
//  - MOSI   - 11 - must use this pin
//  - DC    - 5 - your pin selection
//  - CS    - 4 - your pin selection
//  - Reset - 3 - your pin selection
//
//  Joystick
//  - switch pin to analog pin A5
//  - X axis to analog pin A6
//  - Y axis to analog pin A7
//  - VCC and ground
//
//  Rotary Encoder - can be connected to any two spare pins
//  - Pin 1 to pin A0
//  - Pin 2 (center pin on the EC11) connected to ground
//  - Pin 3 to pin A1
//  - SW pin 1 to digital pin 12 - currently the switch has no functionality in the code
//  - SW pin 2 connected to ground
//
//  SN74LS245 or equivalent - needed between the monitor Uno/Nano transmit pin and the GRBL controller receive pin.  
//  You cannot connect the two together as it will interfere with the GRBL serial tranmission to the PC.
//  - monitor tx pin - input on SN74LS245 - output of SN74LS245 - GRBL controller rx pin
//  - digital pin 7 to pin 19 on SN74LS245 - controls gate
//  - pin 1 of SN74LS245 tied to Vcc
//
//    CD4066
//    - if using this chip you must assert the control line low to turn it off - see the defines below
//
//  Second serial interface using the AltSoftSerial library.
//  Use a USB to serial interface module connected as follows:
//  - the module TX and RX pins connect to the Uno/Nano pins 8 and 9.
//  - AltSoftSerial uses pin 8 as the RX pin and pin 9 as the TX pin.
//
//  Tri-Color LED
//  D6 - Red
//  D7 - Green
//  D10 - Blue

////
//
// Master defines
//
#define testMode 0 //Set to one (1) to ignore state status
#define altSerialEnabled 0 //Set to one (1) to include compiling altSoftSerial
#define debugLevelSystemStateMessaging 0
#define debugLevelMessageProcessing 0
#define debugLevelMachineMovement 0
#define debugLevelProcedureCallTree 0
#define debuglevelSystemFlags 0
#define debuglevelBitFlags 0
#define debuglevelGRBLData 0
#define debuglevelGRBLBadData 0

//Includes
#include <PCD8544.h> //Library version 1.4.3 - note that I added two functions to it - to return the width and height of the display
#if altSerialEnabled
  #include <AltSoftSerial.h>
#endif
#include <SPI.h>
#include <Rotary.h>

//Defines
#define bCNCSPEED 115200
#define altSerialSPEED 38400 //38400 57600 74880 115200
#define txTransmitOnOpenTimeLong 12 //Default amount of time that the Uno/Nano TX pin is allowed to transmit to the GRBL controller - jog movement commands
#define txTransmitOnOpenTimeShort 4 //Default amount of time that the Uno/Nano TX pin is allowed to transmit to the GRBL controller - all other commands
#define txSignalPinOnTime 4 //ms
#define maxJogIncrements 9
#define maxFeedIncrements 6
#define maxZAxisJogIncrements 6
#define maxZAxisFeedIncrements 6
#define maxzSafeIncrements 6
#define maxCmd1Strings 2
#define maxCmd2Strings 2
#define maxCmd3Strings 3
#define menuItemIndent 10
#define menuItemReturnIndent 45
#define displayCNCDataUpdateCounterMax 100 //After this many iterations the display will re-initialize and then continue.  
                                          //Removes garbage that sneaks in...
#define systemPollingPeriod 250 //Amount of milliseconds between polling for more GRBL data.  The display can only update so fast...
#define ledPollingPeriod 250 //500
#define splashScreenDelay 5000 //Amount of time in ms that the splash screen is displayed
#define numChars 128 //Number of elements in the character array
#define notFoundInString -1
#define maxOks 1 //10
#define heartBeatCycleMaxIterations 50
#define noGRBLdataCounterMax 30000
#define zAxisFeedRate 200

//Defines for the tranmist control switch - HW dependent state - see notes above
#define txSwitchOnPin 2 //Pin used to control the on/off gate of the transmit control switch
#define txSwitchStateOn HIGH //Must be adjusted depending upon the type of electronic switch being used - either HIGH or LOW
#define txSwitchStateOff LOW //Must be adjusted depending upon the type of electronic switch being used - either HIGH or LOW

//Signal Pins to GRBL controller
#define resetAbortInterfacePin A2
#define feedHoldInterfacePin A3
#define cycleStartResumeInterfacePin A4

//Joystick Pin Connections and Defines - used for movement in the X and Y axis
#define SW_pin A5 // analog pin connected to switch output of joystick
#define X_pin A6 // analog pin connected to X output of joystick
#define Y_pin A7 // analog pin connected to Y output of joystick
#define yAxisHold 0
#define yAxisFront 1
#define yAxisBack -1
#define xAxisHold 0
#define xAxisLeft -1
#define xAxisRight 1
#define swPressed 0
#define minAnalogRange 0
#define maxAnalogRange 800 //1023

//Rotary Encoder Pin Connections - used for movement in the Z axis
#define rotaryPin1 A0                    
#define rotaryPin2 A1
#define rotarySwitch 12 //currently no functionality in this code

//TriColor LED
#define ledRed 6
#define ledGreen 7
#define ledBlue 10
#define toON HIGH
#define toOFF LOW

//Displays - TFT and 5110
#define dispSCLK 13             
#define dispMOSI 11             
#define dispDC   5
#define dispCS   4
#define dispRST  3  
#define dispTextWidth5110 6 //pixels
#define dispTextHeight 8 //pixels - TFT
#define dispTextSizeMasterScale 2 //Use 1 for small font, use 2 for large font on the TFT display
#define dispTextScale5110 1 //Scale factor - 1 - have not tried scaling font on the 5110 - needs the Adafruit libraries
#define dispRotation 3 // 0 - portrait, 3 - landscape
#define dispTftTextRowHeightPadding 3 //Note you cannot pad a 5110 row height (I think?)
#define dispBaseYPos 5 //pixels - base starting position for the top of the first row
#define cursorGlyph ">" //change to whatever char you want
//
#define colorBlack 0 
#define colorBlue 0
#define colorRed 0
#define colorGreen 0
#define colorCyan 0
#define colorMagenta 0
#define colorYellow 0
#define colorWhite 0
#define colorDarkGreen 0
#define colorGrey 0
#define colorOffWhite 0
#define colorBrown 0
#define colorLightGrey 0
#define colorLightBlue 0
#define colorLightYellow 0
#define colorLightYellow2 0
#define colorDarkYellow 0

#define dispInitFlag 0
#define dispTextWrapT true
#define dispTextWrapF false
//
#define pageTitleColor colorBlack
#define pageBackGroundColor colorLightGrey
#define pageMenuItemColor colorLightBlue
#define pageReturnItemColor colorDarkGreen
#define pageOptionsColor colorGrey
#define pageBCNCStateColor colorRed
#define pageDynamicDataColor colorBrown
#define pageCursorColor colorBlack

#define SYSSTATE_IDLE           0      // Must be zero. No flags set.
#define SYSSTATE_ALARM          bit(0) // In alarm state. Locks out all g-code processes. Allows settings access.
#define SYSSTATE_CHECK_MODE     bit(1) // G-code check mode. Locks out planner and motion only.
#define SYSSTATE_HOMING         bit(2) // Performing homing cycle
#define SYSSTATE_CYCLE          bit(3) // Cycle is running or motions are being executed.
#define SYSSTATE_HOLD           bit(4) // Active feed hold
#define SYSSTATE_JOG            bit(5) // Jogging mode.
#define SYSSTATE_SAFETY_DOOR    bit(6) // Safety door is ajar. Feed holds and de-energizes system.
#define SYSSTATE_SLEEP          bit(7) // Sleep state.

#define SYSFLAGS_NONE           0      // Must be zero. No flags set.
#define SYSFLAGS_NCP            bit(0) // NCP - No connection between this box and GRBL controller.
#define SYSFLAGS_HOMEMSG        bit(1) // HOME message from GRBL received.
#define SYSFLAGS_MALARM         bit(2) // ALARM: message from GRBL.
#define SYSFLAGS_JSPRESSED      bit(3) // JoyStick Button pressed.
#define SYSFLAGS_CMP            bit(4) // Change Menu Page
#define SYSFLAGS_5              bit(5) // 
#define SYSFLAGS_COORDMODE      bit(6) // 0 = MPos, 1 = WPos
#define SYSFLAGS_ALARMACK       bit(7) // 

#define MOVEFLAG_NONE           0      // Must be zero. No flags set.
#define MOVEFLAG_JSL            bit(0) // Left indicator
#define MOVEFLAG_JSR            bit(1) // Right indicator
#define MOVEFLAG_JSF            bit(2) // Front indicator
#define MOVEFLAG_JSB            bit(3) // Back indicator
#define MOVEFLAG_ROTU           bit(4) // Up indicator
#define MOVEFLAG_ROTD           bit(5) // Down indicator
#define MOVEFLAG_XYAXIS         bit(6) // Movement on either the X or Y axis
#define MOVEFLAG_ZAXIS          bit(7) // Movement on the Z axis

#define JOGBLINK_NONE           0      // Must be zero. No flags set.
#define JOGBLINK_L              bit(0) // Left indicator
#define JOGBLINK_R              bit(1) // Right indicator
#define JOGBLINK_F              bit(2) // Front indicator
#define JOGBLINK_B              bit(3) // Back indicator
#define JOGBLINK_Z              bit(4) // Up/Down indicator
#define JOGBLINK_5              bit(5) // 
#define JOGBLINK_6              bit(6) // 
#define JOGBLINK_STATE          bit(7) //

// Bit field and masking macros
#define bit(n) (1 << n)
#define bit_true(x,mask) (x) |= (mask)
#define bit_false(x,mask) (x) &= ~(mask)
#define bit_istrue(x,mask) ((x & mask) != 0)
#define bit_isfalse(x,mask) ((x & mask) == 0)

//Variables
//Use this website to create the Glyph: http://carlosefr.github.io/pcd8544/
//static const byte menuCursor[] = { B00010000, B00110100, B00110000, B00110100, B00010000 }; //This is a bitmap of a smiley - used to mark the menu cursor position
//static const byte menuCursor[] = {0x10, 0x10, 0x54, 0x38, 0x10}; //This is a bitmap of an arrow - used to mark the menu cursor position

//State Flags - bitsets
uint8_t sysState;
uint8_t sysFlags;
uint8_t movFlags;
uint8_t jogBlinkFlags;

//Vars
float fxPos = 0;
float fyPos = 0;
float fzPos = 0;
float fwcoXPos = 0;
float fwcoYPos = 0;
float fwcoZPos = 0;
String pinStates = "";
String movementCommandString = "";
String movementZAxisCommandString = "";
byte blinkState = 0;
int feedRate = 0;
int spindleRate = 0;
int lastYAxisDir = yAxisHold;
int lastXAxisDir = xAxisHold;
unsigned char lastZMovementDir = DIR_NONE;
boolean sendJogCancel;

//Counters and counts
int noGRBLdataConnectionCounter = 0;
byte displayCNCDataUpdateCounter = 0; 
byte heartBeatCycleCount = 0;
byte pinStatesLength = 0;

//Menu Enumumerators - used to manage the current menu page and selected item on the displayed page
enum menuPagesOptionsVal {Main_L1, cncOptions_L1, Jogging_L1, Cutting_L1, Commands_L1, WPos_L1, NCP_Alarm_L1, MAS_Alarm_L1, HOME_Alarm_L1};
enum menuPagesOptionsVal menuPageActive = Main_L1; //Start with displaying the main menu page
//
enum mainMenuOptionsVal {mmCommands, mmCncOptions, mmJogging, mmWPos, mmCutting};
enum mainMenuOptionsVal mainMenuItemSelected = mmCncOptions; //Default the selected menu to: mmCncOptions
enum mainMenuOptionsVal previousMainMenuItemSelected;
//
enum jogMenuOptionsVal {jmZIncrement, jmZFeed, jmXYIncrement, jmXYFeed, jmReturn};
enum jogMenuOptionsVal jogMenuItemSelected = jmReturn; //Default the selected menu to: jmReturn
enum jogMenuOptionsVal previousJogMenuItemSelected;
//
enum commandsMenuOptionsVal {coHome, coSleepWake, coHoldResume, coAbortResetUnlock, coReturn};
enum commandsMenuOptionsVal commandsMenuItemSelected = coHome; //Default the selected menu to: coHome
enum commandsMenuOptionsVal previousCommandsMenuItemSelected;
//
enum wPosMenuOptionsVal {wpSetXYzero, wpSetZzero, wpHome, wpZSafe, wpReturn};
enum wPosMenuOptionsVal wPosMenuItemSelected = wpHome; //Default the selected menu to: wpHome
enum wPosMenuOptionsVal previousWPosMenuItemSelected;

//Display enumerators
enum textPosition {text_XCoord_YCoord, text_XLeftJustified_YCoord, text_XCenterJustified_YCoord, text_XRightJustified_YCoord, text_XCoord_YLineJustified, text_XLeftJustified_YLineJustified, text_XCenterJustified_YLineJustified, text_XRightJustified_YLineJustified};
enum textPosition textPositionFormat;

//Jogging selectable values in the Jog Options menu - Distance Increment and Feed rate (speed) increment.  
//This program us using milimeters as the measurement (G21) in the command syntax.

byte jogIncrement[maxJogIncrements] = {1, 5, 10, 15, 20, 30, 40, 50, 100}; //index starts at 0 - these represent mm of distance
byte jogIncrementIndex = 7; //default 7
int feedIncrement[maxFeedIncrements] = {50, 100, 200, 400, 600, 800}; //index starts at 0
byte feedIncrementIndex = 5; //default 5
//
float jogZAxisIncrement[maxZAxisJogIncrements] = {0.1, 0.5, 1, 5, 10, 20}; //index starts at 0 - these represent mm of distance
byte jogZAxisIncrementIndex = 3; //default 3
int feedZAxisIncrement[maxZAxisFeedIncrements] = {10, 30, 50, 100, 300, 500}; //index starts at 0
byte feedZAxisIncrementIndex = 4; //default 4
int zSafeIncrement[maxzSafeIncrements] = {10, 15, 20, 25, 30, 40}; //index starts at 0
byte zSafeIncrementIndex = 3;

//CMds selectable values in some menu item
String cmd1Strings[maxCmd1Strings] = {"Sleep.", ".Wake Up"}; //Note: index starts at 0
byte cmd1StringsIndex = 0;

//CMds selectable values in some menu item
String cmd2Strings[maxCmd2Strings] = {"Hold.", ".Resume"}; //Note: index starts at 0
byte cmd2StringsIndex = 0;

//CMds selectable values in some menu item
String cmd3Strings[maxCmd3Strings] = {"Abort..", ".Reset.", "..Unlock"}; //Note: index starts at 0
byte cmd3StringsIndex = 0;

//GRBL Commands
char cmdJogCancel = 0x85; //Jog Cancel
char cmdSoftReset = 0x18; //Soft Reset (Ctrl-x)
String cmdSleepMode = "$SLP\n"; //Enable Sleep Mode
String cmdHomingCycle = "$H\n"; //Run Homing Cycle
String cmdKillAlarmLock = "$X\n"; //Kill Alarm Lock
//String cmdStatusReportQuery = "?\n";  //Status Report Query  -- NOT USED - implemented with signal pin
//String cmdCycleStartResume = "~\n";  //Cycle Start/Resume  -- NOT USED - implemented with signal pin
//String cmdFeedHold = "!\n"; //Feed Hold  -- NOT USED - implemented with signal pin

//Timers
unsigned long systemPollingTimer = millis();
unsigned long ledPollingTimer = millis();

//Objects
#if altSerialEnabled
  AltSoftSerial altSerial;
#endif
Rotary zAxis = Rotary(rotaryPin1, rotaryPin2); //Rotary encoder for moving the Z Axis up and down
static PCD8544 lcd5110Display; //Basic library to drive the Nokia 5110 display

////
//
// ** SETUP **
//
////
void setup() {
  //
  initSerialInterface(); //Serial Interfaces
  initJoyStickPins(); //Joystick setup
  initRotaryPins(); //Rotary Setup
  initGRBLPins(); //GRBL Interface H/W pins
  initLEDPins(); //LED pins
  initHWTxSwitchPin(); //HW Tx switch pin
  // Initialize the display and cycle LED color scheme indicating system initializing
  displayInit(dispCS, dispDC, dispRST, dispInitFlag, dispRotation); //Sets (csPin, dcPin, rstPin, displayType, rotationPoint)
  displaySplashScreen(true, dispTextWrapF, colorDarkYellow); //Display the Splash Screen
  displayMainMenuOptions(true, mainMenuItemSelected, mainMenuItemSelected, true); //Start off with the main menu page - default menu item is defined in the Enums above.
  cycleLEDs(250); 
  //Set up global state and flags
  sysState = SYSSTATE_IDLE;
  sysFlags = SYSFLAGS_NONE;
  movFlags = MOVEFLAG_NONE;
  jogBlinkFlags = JOGBLINK_NONE;
  bit_false(sysFlags, SYSFLAGS_COORDMODE); //default the coordinate system 0(false) = MPos, 1(true) = WPos //use bit_true for WPos
  #if debugLevelProcedureCallTree
  altSerial.println(F("**** Done Setup ****"));
  #endif
}

////
//
// ** MAINLINE **
//
////
void loop() {
  //
  #if altSerialEnabled
    //This would be used if you have another serial connection as defined above and enter some text and send it to the Uno/Nano from a terminal like PUTTY.
    if (altSerial.available()) {readAltSerialMessage();}  
  #endif
  if (Serial.available()) {//This is the serial connection from GRBL to this Uno/Nano.  Process the incoming data feed from GRBL.
    readGRBLDataStream();
  } else { //We are not getting any GRBL data feed anymore - could be a no connection problem (NCP) event
    processNoGRBLDataStream();
  }
  
  //If we have a Z axis movement to process, send it to GRBL
  if (bit_istrue(movFlags, MOVEFLAG_ZAXIS) && (movementZAxisCommandString != "")) {
    sendMessageToGRBL(txTransmitOnOpenTimeLong, movementZAxisCommandString); movementZAxisCommandString = ""; bit_false(movFlags, MOVEFLAG_ZAXIS); 
  }
  //If we have a X or Y axis movement to process, send it to GRBL
  if (bit_istrue(movFlags, MOVEFLAG_XYAXIS) && (movementCommandString != "")) {
    sendMessageToGRBL(txTransmitOnOpenTimeLong, movementCommandString); movementCommandString = ""; bit_false(movFlags, MOVEFLAG_XYAXIS); \
  }
  
  //Process the Rotary encoder - controls the Z axis movement up and down and the coordinate system selection when on the cutting page
  readZRotaryInput(); 
  processZAxisMovement();
  
  //Update the System Inputs based upon the polling timer
  if (millis() - systemPollingTimer > systemPollingPeriod) { //Only do the updates every polling time period.  The systemPollingPeriod is defined above.       
    readJoystickInputs(); //Process the joystick inputs in the X and Y axis as well as the button push
    //Manage the state machine and update the displayed menu page
    manageStateMachine();
    updatelcd5110Displays();
    //Manage the menu cursor movements related to the joystick inputs and button press 
    manageMenuJSMovements();
    manageMenuJSButtonPress();
    manageMenuJSParamsSelect();
    systemPollingTimer = millis(); //Reset the polling timer for the next cycle         
    if (heartBeatCycleCount > heartBeatCycleMaxIterations) { //Send out the heart beat message to Altserial
      heartBeatCycleCount = 0;
      #if altSerialEnabled
      altSerial.println(F("Heartbeat!"));
      #endif
    } else {
      heartBeatCycleCount++;
    }
  }
}
